# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base interface for searchable embeddings databases."""

import abc
import collections
from collections.abc import Sequence
import datetime as dt
from typing import Any, Literal

from ml_collections import config_dict
import numpy as np
from perch_hoplite.db import datatypes
from perch_hoplite.db import score_functions
from perch_hoplite.db import search_results


class HopliteDBInterface(abc.ABC):
  """Interface for searchable embeddings database with metadata.

  The database consists of a table of embedding windows, with a unique id for
  each window and some additional tables for linking the embedding window to
  recording-level and site-level metadata. Additionally, a key-value table of
  ConfigDict objects is used to store Hoplite-specific metadata associated with
  the database.

  The interface provides multiple methods for filtering metadata, each of them
  accepting one or more ConfigDict filters with the following structure:

    - eq => to test if given column is equal to given value
      - column1: value1
      - column2: None
    - neq => to test if given column is not equal to given value
      - column1: value1
      - column2: None
    - lt => to test if given column is less than given value
      - column: value
    - lte => to test if given column is less than or equal to given value
      - column: value
    - gt => to test if given column is greater than given value
      - column: value
    - gte => to test if given column is greater than or equal to given value
      - column: value
    - isin => to test if given column is in given list of values
      - column: [value1, value2, value3, ...]
    - notin => to test if given column is not in given list of values
      - column: [value1, value2, value3, ...]
    - range => to test if given column is between two values
      - column: [value1, value2]
    - approx => to test if given column is approximately equal to given value
                (within 1e-6 difference); useful for floating point comparisons
      - column: value

  The recommended way to build such ConfigDict filters is to use something like
  this (feel free to omit operations that are not needed):

  ```python
  from ml_collections import config_dict
  filter_dict = config_dict.create(
      eq=dict(column1=value1, column2=None),
      neq=dict(column1=value1, column2=None),
      lt=dict(column=value),
      lte=dict(column=value),
      gt=dict(column=value),
      gte=dict(column=value),
      isin=dict(column=[value1, value2, value3]),
      notin=dict(column=[value1, value2, value3]),
      range=dict(column=[value1, value2]),
      approx=dict(column=value),
  )
  ```
  """

  @classmethod
  @abc.abstractmethod
  def create(cls, **kwargs) -> "HopliteDBInterface":
    """Connect to and, if needed, initialize the database.

    Args:
      **kwargs: Keyword arguments to pass to the implementation-specific
        `create` function.

    Returns:
      A new instance of the database.
    """

  @abc.abstractmethod
  def add_extra_table_column(
      self,
      table_name: str,
      column_name: str,
      column_type: type[Any],
  ) -> None:
    """Add an extra column to a table in the database.

    Args:
      table_name: The name of the table to add the column to.
      column_name: The name of the column to add.
      column_type: The type of the column to add.
    """

  @abc.abstractmethod
  def get_extra_table_columns(self) -> dict[str, dict[str, type[Any]]]:
    """Get all extra columns in the database."""

  @abc.abstractmethod
  def commit(self) -> None:
    """Commit any pending transactions to the database."""

  @abc.abstractmethod
  def rollback(self) -> None:
    """Rollback any pending transactions to the database."""

  @abc.abstractmethod
  def thread_split(self) -> "HopliteDBInterface":
    """Get a new instance of the database with the same contents.

    For example, SQLite databases need a distinct object in each thread.

    Returns:
      A new instance of the database.
    """

  @abc.abstractmethod
  def insert_metadata(self, key: str, value: config_dict.ConfigDict) -> None:
    """Insert a key-value pair into the metadata table.

    Args:
      key: String for metadata key.
      value: ConfigDict object to store.
    """

  @abc.abstractmethod
  def get_metadata(self, key: str | None) -> config_dict.ConfigDict:
    """Get a key-value pair from the metadata table.

    Args:
      key: String for metadata key to retrieve. If None, returns all metadata.

    Returns:
      A ConfigDict containing the metadata.

    Raises:
      KeyError: If `key` is not found in the metadata table.
    """

  @abc.abstractmethod
  def remove_metadata(self, key: str | None) -> None:
    """Remove a key-value pair from the metadata table.

    Args:
      key: String for metadata key to remove. If None, removes all metadata.

    Raises:
      KeyError: If `key` is not found in the metadata table.
    """

  @abc.abstractmethod
  def insert_deployment(
      self,
      name: str,
      project: str,
      latitude: float | None = None,
      longitude: float | None = None,
      **kwargs: Any,
  ) -> int:
    """Insert a deployment into the database.

    Args:
      name: The name of the deployment.
      project: The project of the deployment.
      latitude: The (optional) latitude of the deployment.
      longitude: The (optional) longitude of the deployment.
      **kwargs: Additional keyword arguments to pass to the deployment.

    Returns:
      The ID of the inserted deployment.
    """

  @abc.abstractmethod
  def get_deployment(self, deployment_id: int) -> datatypes.Deployment:
    """Get a deployment from the database.

    Args:
      deployment_id: The ID of the deployment to retrieve.

    Returns:
      A Deployment object containing the requested information.
    """

  @abc.abstractmethod
  def remove_deployment(self, deployment_id: int) -> None:
    """Remove a deployment from the database.

    Args:
      deployment_id: The ID of the deployment to remove.
    """

  @abc.abstractmethod
  def insert_recording(
      self,
      filename: str,
      datetime: dt.datetime | None = None,
      deployment_id: int | None = None,
      **kwargs: Any,
  ) -> int:
    """Insert a recording into the database.

    Args:
      filename: The filename of the recording.
      datetime: The (optional) datetime of the recording.
      deployment_id: The (optional) ID of the deployment to which the recording
        belongs.
      **kwargs: Additional keyword arguments to pass to the recording.

    Returns:
      The ID of the inserted recording.
    """

  @abc.abstractmethod
  def get_recording(self, recording_id: int) -> datatypes.Recording:
    """Get a recording from the database.

    Args:
      recording_id: The ID of the recording to retrieve.

    Returns:
      A Recording object containing the requested information.
    """

  @abc.abstractmethod
  def remove_recording(self, recording_id: int) -> None:
    """Remove a recording from the database.

    Args:
      recording_id: The ID of the recording to remove.
    """

  def _handle_window_duplicates(
      self,
      recording_id: int,
      offsets: list[float],
      handle_duplicates: Literal[
          "allow", "overwrite", "skip", "error"
      ] = "error",
  ) -> int | None:
    """Utility function to handle window duplicates in the database.

    Args:
      recording_id: The recording ID of potential window duplicates.
      offsets: The offsets of potential window duplicates.
      handle_duplicates: How to handle entries matching another database item's
        (recording_id, offsets). If "allow", duplicates are left as they are and
        `None` is returned. If "overwrite", duplicates are removed and `None` is
        returned. If "skip", the ID of the first matching window is returned if
        duplicates are found, otherwise `None` is returned. If "error", an error
        is raised if duplicates are found, otherwise `None` is returned (this is
        the default).

    Returns:
      The ID of the first matching window if `handle_duplicates` is "skip" and
      duplicates are found, otherwise `None`.

    Raises:
      RuntimeError: If `handle_duplicates` is "error" and duplicates are found.
    """

    if handle_duplicates in ["overwrite", "skip", "error"]:
      matches = self.get_all_windows(
          filter=config_dict.create(
              eq=dict(recording_id=recording_id),
              approx=dict(offsets=offsets),
          )
      )
      if matches:
        if handle_duplicates == "overwrite":
          for match in matches:
            self.remove_window(match.id)
        elif handle_duplicates == "skip":
          return matches[0].id
        elif handle_duplicates == "error":
          raise RuntimeError(
              f"Duplicate window found (id = {matches[0].id}), but"
              ' `handle_duplicates` is set to "error".'
          )

    return None

  @abc.abstractmethod
  def insert_window(
      self,
      recording_id: int,
      offsets: list[float],
      embedding: np.ndarray | None = None,
      handle_duplicates: Literal[
          "allow", "overwrite", "skip", "error"
      ] = "error",
      **kwargs: Any,
  ) -> int:
    """Insert a window into the database.

    Args:
      recording_id: The ID of the recording to which the window belongs.
      offsets: The offsets of the window.
      embedding: The embedding vector. If None, no embedding vector is inserted
        into the database for this particular window.
      handle_duplicates: How to handle entries matching another database item's
        (recording_id, offsets). If "allow", duplicates are allowed. If
        "overwrite", the new window overwrites the old one. If "skip", the old
        window is kept. If "error", an error is raised if duplicates are found
        (this is the default).
      **kwargs: Additional keyword arguments to pass to the window.

    Returns:
      The ID of the inserted window (if `handle_duplicates` is "allow" or
      "overwrite"), or the ID of the first matching window (if
      `handle_duplicates` is "skip").

    Raises:
      RuntimeError: If `handle_duplicates` is "error" and duplicates are found.
    """

  def insert_windows_batch(
      self,
      windows_batch: Sequence[dict[str, Any]],
      embeddings_batch: np.ndarray | None = None,
      handle_duplicates: Literal[
          "allow", "overwrite", "skip", "error"
      ] = "error",
  ) -> Sequence[int]:
    """Insert a batch of windows into the database.

    Args:
      windows_batch: A sequence of windows to insert. Each window must be a dict
        with same keys as the arguments of `insert_window()`, except for the
        `embedding` argument.
      embeddings_batch: A batch of embedding vectors for the given windows. If
        None, no embedding vectors are inserted into the database.
      handle_duplicates: How to handle entries matching another database item's
        (recording_id, offsets). If "allow", duplicates are allowed. If
        "overwrite", the new window overwrites the old one. If "skip", the old
        window is kept. If "error", an error is raised if duplicates are found
        (this is the default).

    Returns:
      A sequence of IDs of inserted and/or matching windows, determined by
      `handle_duplicates`. See the return value of `insert_window()` for more
      details.

    Raises:
      RuntimeError: If `handle_duplicates` is "error" and duplicates are found,
        or if `handle_duplicates` is not "allow" and there are duplicates in
        `windows_batch` itself.
    """

    # Make sure that, unless we're in "allow" mode, there are no duplicates in
    # the batch itself.
    if handle_duplicates != "allow":
      for i in range(len(windows_batch)):
        for j in range(i + 1, len(windows_batch)):
          if windows_batch[i]["recording_id"] == windows_batch[j][
              "recording_id"
          ] and np.allclose(
              windows_batch[i]["offsets"],
              windows_batch[j]["offsets"],
              rtol=0.0,
              atol=1e-6,
          ):
            raise RuntimeError(
                "Duplicates found in `windows_batch`, but this use case is not"
                ' supported unless `handle_duplicates` is set to "allow"'
                f' (handle_duplicates = "{handle_duplicates}").'
            )

    # Insert the windows one by one, in the order they are given.
    window_ids = []
    for window_kwargs, embedding in zip(windows_batch, embeddings_batch):
      window_id = self.insert_window(
          embedding=embedding,
          handle_duplicates=handle_duplicates,
          **window_kwargs,
      )
      window_ids.append(window_id)
    return window_ids

  @abc.abstractmethod
  def get_window(
      self,
      window_id: int,
      include_embedding: bool = False,
  ) -> datatypes.Window:
    """Get a window from the database.

    Args:
      window_id: The ID of the window to retrieve.
      include_embedding: Whether to include the embedding vector in the returned
        Window object.

    Returns:
      A Window object containing the requested information.
    """

  @abc.abstractmethod
  def get_embedding(self, window_id: int) -> np.ndarray:
    """Get an embedding vector from the database.

    Args:
      window_id: The window ID of the embedding to retrieve.

    Returns:
      An embedding vector for the given window ID.
    """

  def get_embeddings_batch(self, window_ids: Sequence[int]) -> np.ndarray:
    """Get a batch of embedding vectors from the database.

    Args:
      window_ids: The window IDs of the embeddings to retrieve.

    Returns:
      A batch of embedding vectors for the given window IDs.
    """

    embeddings = [self.get_embedding(window_id) for window_id in window_ids]
    return np.stack(embeddings)

  @abc.abstractmethod
  def remove_window(self, window_id: int) -> None:
    """Remove a window from the database.

    Args:
      window_id: The ID of the window to remove.
    """

  def _handle_annotation_duplicates(
      self,
      recording_id: int,
      offsets: list[float],
      label: str,
      label_type: datatypes.LabelType,
      provenance: str,
      handle_duplicates: Literal[
          "allow", "overwrite", "skip", "error"
      ] = "error",
  ) -> int | None:
    """Utility function to handle annotation duplicates in the database.

    Args:
      recording_id: The recording ID of potential annotation duplicates.
      offsets: The offsets of potential annotation duplicates.
      label: The label of potential annotation duplicates.
      label_type: The label type of potential annotation duplicates.
      provenance: The provenance of potential annotation duplicates.
      handle_duplicates: How to handle entries matching another database item's
        (recording_id, offsets, label, label_type, provenance). If "allow",
        duplicates are left as they are and `None` is returned. If "overwrite",
        duplicates are removed and `None` is returned. If "skip", the ID of the
        first matching annotation is returned if duplicates are found, otherwise
        `None` is returned. If "error", an error is raised if duplicates are
        found, otherwise `None` is returned (this is the default).

    Returns:
      The ID of the first matching annotation if `handle_duplicates` is "skip"
      and duplicates are found, otherwise `None`.

    Raises:
      RuntimeError: If `handle_duplicates` is "error" and duplicates are found.
    """

    if handle_duplicates in ["overwrite", "skip", "error"]:
      matches = self.get_all_annotations(
          filter=config_dict.create(
              eq=dict(
                  recording_id=recording_id,
                  label=label,
                  label_type=label_type,
                  provenance=provenance,
              ),
              approx=dict(
                  offsets=offsets,
              ),
          )
      )
      if matches:
        if handle_duplicates == "overwrite":
          for match in matches:
            self.remove_annotation(match.id)
        elif handle_duplicates == "skip":
          return matches[0].id
        elif handle_duplicates == "error":
          raise RuntimeError(
              f"Duplicate annotation found (id = {matches[0].id}), but"
              ' `handle_duplicates` is set to "error".'
          )

    return None

  @abc.abstractmethod
  def insert_annotation(
      self,
      recording_id: int,
      offsets: list[float],
      label: str,
      label_type: datatypes.LabelType,
      provenance: str,
      handle_duplicates: Literal[
          "allow", "overwrite", "skip", "error"
      ] = "error",
      **kwargs: Any,
  ) -> int:
    """Insert an annotation into the database.

    Args:
      recording_id: The ID of the recording to which the annotation refers to.
      offsets: The offsets in the recording to which the annotation refers to.
      label: The annotation label.
      label_type: The type of label (e.g. positive or negative).
      provenance: The provenance of the annotation.
      handle_duplicates: How to handle entries matching another database item's
        (recording_id, offsets, label, label_type, provenance). If "allow",
        duplicates are allowed. If "overwrite", the new annotation overwrites
        the old one. If "skip", the old annotation is kept. If "error", an error
        is raised if duplicates are found (this is the default).
      **kwargs: Additional keyword arguments to pass to the annotation.

    Returns:
      The ID of the inserted annotation (if `handle_duplicates` is "allow" or
      "overwrite"), or the ID of the first matching annotation (if
      `handle_duplicates` is "skip").

    Raises:
      RuntimeError: If `handle_duplicates` is "error" and duplicates are found.
    """

  @abc.abstractmethod
  def get_annotation(self, annotation_id: int) -> datatypes.Annotation:
    """Get an annotation from the database.

    Args:
      annotation_id: The ID of the annotation to retrieve.

    Returns:
      An Annotation object containing the requested information.
    """

  @abc.abstractmethod
  def remove_annotation(self, annotation_id: int) -> None:
    """Remove an annotation from the database.

    Args:
      annotation_id: The ID of the annotation to remove.
    """

  @abc.abstractmethod
  def count_embeddings(self) -> int:
    """Get the number of embeddings in the database."""

  @abc.abstractmethod
  def match_window_ids(
      self,
      deployments_filter: config_dict.ConfigDict | None = None,
      recordings_filter: config_dict.ConfigDict | None = None,
      windows_filter: config_dict.ConfigDict | None = None,
      annotations_filter: config_dict.ConfigDict | None = None,
      limit: int | None = None,
  ) -> Sequence[int]:
    """Get matching window IDs from the database based on given filters.

    Args:
      deployments_filter: If provided, only retrieve window IDs that have
        deployments matching constraints specified by this filter.
      recordings_filter: If provided, only retrieve window IDs that have
        recordings matching constraints specified by this filter.
      windows_filter: If provided, only retrieve window IDs that have windows
        matching constraints specified by this filter.
      annotations_filter: If provided, only retrieve window IDs that have
        annotations matching constraints specified by this filter.
      limit: If provided, limit the number of window IDs returned.

    Returns:
      A sequence of window IDs from the database.
    """

  @abc.abstractmethod
  def get_all_projects(self) -> Sequence[str]:
    """Get all distinct projects from the database.

    Returns:
      A sequence of all projects in the database.
    """

  @abc.abstractmethod
  def get_all_deployments(
      self,
      filter: config_dict.ConfigDict | None = None,  # pylint: disable=redefined-builtin
  ) -> Sequence[datatypes.Deployment]:
    """Get all deployments from the database.

    Args:
      filter: If provided, only retrieve deployments matching constraints
        specified by this filter.

    Returns:
      A sequence of all matching deployments in the database.
    """

  @abc.abstractmethod
  def get_all_recordings(
      self,
      filter: config_dict.ConfigDict | None = None,  # pylint: disable=redefined-builtin
  ) -> Sequence[datatypes.Recording]:
    """Get all recordings from the database.

    Args:
      filter: If provided, only retrieve recordings matching constraints
        specified by this filter.

    Returns:
      A sequence of all matching recordings in the database.
    """

  @abc.abstractmethod
  def get_all_windows(
      self,
      include_embedding: bool = False,
      filter: config_dict.ConfigDict | None = None,  # pylint: disable=redefined-builtin
  ) -> Sequence[datatypes.Window]:
    """Get all windows from the database.

    Args:
      include_embedding: Whether to include the embedding vector in the returned
        Window objects.
      filter: If provided, only retrieve windows matching constraints specified
        by this filter.

    Returns:
      A sequence of all matching windows in the database.
    """

  @abc.abstractmethod
  def get_all_annotations(
      self,
      filter: config_dict.ConfigDict | None = None,  # pylint: disable=redefined-builtin
  ) -> Sequence[datatypes.Annotation]:
    """Get all annotations from the database.

    Args:
      filter: If provided, only retrieve annotations matching constraints
        specified by this filter.

    Returns:
      A sequence of all matching annotations in the database.
    """

  @abc.abstractmethod
  def get_all_labels(
      self,
      label_type: datatypes.LabelType | None = None,
  ) -> Sequence[str]:
    """Get all distinct labels from the database.

    Args:
      label_type: If provided, filters to the target label type.

    Returns:
      A sequence of all labels in the database.
    """

  @abc.abstractmethod
  def count_each_label(
      self,
      label_type: datatypes.LabelType | None = None,
  ) -> collections.Counter[str]:
    """Count each label in the database, ignoring provenance.

    Args:
      label_type: If provided, filters to the target label type.

    Returns:
      A counter dict for labels.
    """

  @abc.abstractmethod
  def get_embedding_dim(self) -> int:
    """Get the embedding dimension."""

  @abc.abstractmethod
  def get_embedding_dtype(self) -> type[Any]:
    """Get the embedding data type."""

  def search(
      self,
      query_embedding: np.ndarray,
      search_list_size: int,
      approximate: bool = True,  # pylint: disable=unused-argument
      target_score: float | None = None,
      score_fn_name: str = "dot",
      **kwargs: Any,
  ) -> search_results.TopKSearchResults:
    """Search for neighbors of the query embedding.

    Args:
      query_embedding: The query embedding vector.
      search_list_size: The number of results to return.
      approximate: Whether to use approximate search.
      target_score: If set, search for examples near this score.
      score_fn_name: The name of the score function to use. Likely ignored if
        `approximate` is True, as the underlying indexed score function will be
        used.
      **kwargs: Additional keyword arguments to pass to the search function.

    Returns:
      A TopKSearchResults object containing the search results.
    """
    # pylint: disable-next=g-import-not-at-top
    from perch_hoplite.db import brutalism

    # Default implementation is brute-force.
    # Note: Implementation should override this if approximate=True.
    if approximate:
      raise NotImplementedError(
          "Approximate search is not implemented for this database."
      )
    score_fn = score_functions.get_score_fn(
        score_fn_name, target_score=target_score
    )
    results = brutalism.threaded_brute_search(
        self,
        query_embedding,
        search_list_size,
        score_fn=score_fn,
        **kwargs,
    )

    if target_score is not None:
      raw_score_fn = score_functions.get_score_fn(
          score_fn_name, target_score=None
      )
      # If target_sampling, the results should be returned with the original
      # (un-targeted) score.
      results = brutalism.rerank(query_embedding, results, self, raw_score_fn)
    return results



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions for training and applying a linear classifier."""

import base64
from collections.abc import Iterator, Sequence
from concurrent import futures
import csv
import dataclasses
import json
from typing import Any

from etils import epath
from ml_collections import config_dict
import numpy as np
from perch_hoplite.agile import classifier_data
from perch_hoplite.agile import metrics
from perch_hoplite.db import interface as db_interface
import tensorflow as tf
import tqdm


@dataclasses.dataclass
class LinearClassifier:
  """Wrapper for linear classifier params and metadata."""

  beta: np.ndarray
  beta_bias: np.ndarray
  classes: tuple[str, ...]
  embedding_model_config: Any

  def __call__(self, embeddings: np.ndarray):
    return np.dot(embeddings, self.beta) + self.beta_bias

  def to_config_dict(self) -> config_dict.ConfigDict:
    """Convert the classifier to a ConfigDict."""
    cfg = config_dict.ConfigDict()
    cfg.model_config = self.embedding_model_config
    cfg.classes = self.classes
    # Convert numpy arrays to base64 encoded blobs.
    beta_bytes = base64.b64encode(np.float32(self.beta).tobytes()).decode(
        'ascii'
    )
    beta_bias_bytes = base64.b64encode(
        np.float32(self.beta_bias).tobytes()
    ).decode('ascii')
    cfg.beta = beta_bytes
    cfg.beta_bias = beta_bias_bytes
    return cfg

  @classmethod
  def from_config_dict(cls, cfg: config_dict.ConfigDict) -> 'LinearClassifier':
    """Create a classifier from a ConfigDict."""
    classes = cfg.classes
    beta = np.frombuffer(base64.b64decode(cfg.beta), dtype=np.float32)
    beta = np.reshape(beta, (-1, len(classes)))
    beta_bias = np.frombuffer(base64.b64decode(cfg.beta_bias), dtype=np.float32)
    embedding_model_config = cfg.model_config
    return cls(beta, beta_bias, classes, embedding_model_config)

  def save(self, path: str):
    """Save the classifier to a path."""
    with open(path, 'w') as f:
      f.write(self.to_config_dict().to_json())

  @classmethod
  def load(cls, path: str):
    """Load a classifier from a path."""
    with open(path, 'r') as f:
      cfg_json = json.loads(f.read())
      cfg = config_dict.ConfigDict(cfg_json)
      return cls.from_config_dict(cfg)


def get_linear_model(embedding_dim: int, num_classes: int) -> tf.keras.Model:
  """Create a simple linear Keras model."""
  model = tf.keras.Sequential([
      tf.keras.Input(shape=[embedding_dim]),
      tf.keras.layers.Dense(num_classes),
  ])
  return model


def bce_loss(
    y_true: tf.Tensor,
    logits: tf.Tensor,
    is_labeled_mask: tf.Tensor,
    weak_neg_weight: float,
) -> tf.Tensor:
  """Binary cross entropy loss from logits with weak negative weights."""
  y_true = tf.cast(y_true, dtype=logits.dtype)
  log_p = tf.math.log_sigmoid(logits)
  log_not_p = tf.math.log_sigmoid(-logits)
  # optax sigmoid_binary_cross_entropy:
  # -labels * log_p - (1.0 - labels) * log_not_p
  raw_bce = -y_true * log_p - (1.0 - y_true) * log_not_p
  is_labeled_mask = tf.cast(is_labeled_mask, dtype=logits.dtype)
  weights = (1.0 - is_labeled_mask) * weak_neg_weight + is_labeled_mask
  return tf.reduce_mean(raw_bce * weights)


def hinge_loss(
    y_true: tf.Tensor,
    logits: tf.Tensor,
    is_labeled_mask: tf.Tensor,
    weak_neg_weight: float,
) -> tf.Tensor:
  """Weighted SVM hinge loss."""
  # Convert multihot to +/- 1 labels.
  y_true = 2 * y_true - 1
  weights = (1.0 - is_labeled_mask) * weak_neg_weight + is_labeled_mask
  raw_hinge_loss = tf.maximum(0, 1 - y_true * logits)
  return tf.reduce_mean(raw_hinge_loss * weights)


def infer(params, embeddings: np.ndarray):
  """Apply the model to embeddings."""
  return np.dot(embeddings, params['beta']) + params['beta_bias']


def eval_classifier(
    params: Any,
    data_manager: classifier_data.DataManager,
    eval_ids: np.ndarray,
) -> dict[str, float]:
  """Evaluate a classifier on a set of examples."""
  iter_ = data_manager.batched_example_iterator(
      eval_ids, add_weak_negatives=False, repeat=False
  )
  # The embedding ids may be shuffled by the iterator, so we will track the ids
  # of the examples we are evaluating.
  got_ids = []
  pred_logits = []
  true_labels = []
  for batch in iter_:
    pred_logits.append(infer(params, batch.embedding))
    true_labels.append(batch.multihot)
    got_ids.append(batch.idx)
  pred_logits = np.concatenate(pred_logits, axis=0)
  true_labels = np.concatenate(true_labels, axis=0)
  got_ids = np.concatenate(got_ids, axis=0)

  # Compute the top1 accuracy on examples with at least one label.
  labeled_locs = np.where(true_labels.sum(axis=1) > 0)
  top_preds = np.argmax(pred_logits, axis=1)
  top1 = true_labels[np.arange(top_preds.shape[0]), top_preds]
  top1 = top1[labeled_locs].mean()

  rocs = metrics.roc_auc(
      logits=pred_logits, labels=true_labels, sample_threshold=1
  )
  cmaps = metrics.cmap(
      logits=pred_logits, labels=true_labels, sample_threshold=1
  )
  return {
      'top1_acc': top1,
      'roc_auc': rocs['macro'],
      'roc_auc_individual': rocs['individual'],
      'cmap': cmaps['macro'],
      'cmap_individual': cmaps['individual'],
      'eval_ids': got_ids,
      'eval_preds': pred_logits,
      'eval_labels': true_labels,
  }


def train_linear_classifier(
    data_manager: classifier_data.DataManager,
    learning_rate: float,
    weak_neg_weight: float,
    num_train_steps: int,
    loss: str = 'bce',
) -> tuple[LinearClassifier, dict[str, float]]:
  """Train a linear classifier."""
  embedding_dim = data_manager.db.get_embedding_dim()
  num_classes = len(data_manager.get_target_labels())
  lin_model = get_linear_model(embedding_dim, num_classes)
  optimizer = tf.keras.optimizers.Adam(learning_rate=learning_rate)
  lin_model.compile(optimizer=optimizer, loss='binary_crossentropy')
  if loss == 'hinge':
    loss_fn = hinge_loss
  elif loss == 'bce':
    loss_fn = bce_loss
  else:
    raise ValueError(f'Unknown loss: {loss}')

  @tf.function
  def train_step(y_true, embeddings, is_labeled_mask):
    with tf.GradientTape() as tape:
      logits = lin_model(embeddings, training=True)
      loss = loss_fn(y_true, logits, is_labeled_mask, weak_neg_weight)
      loss = tf.reduce_mean(loss)
    grads = tape.gradient(loss, lin_model.trainable_variables)
    optimizer.apply_gradients(zip(grads, lin_model.trainable_variables))
    return loss

  train_idxes, eval_idxes = data_manager.get_train_test_split()
  train_iter_ = data_manager.batched_example_iterator(
      train_idxes, add_weak_negatives=True, repeat=True
  )
  progress = tqdm.tqdm(enumerate(train_iter_), total=num_train_steps)

  for step, batch in enumerate(train_iter_):
    if step >= num_train_steps:
      break
    step_loss = train_step(
        batch.multihot, batch.embedding, batch.is_labeled_mask
    )
    progress.update(n=1)
    progress.set_description(f'Loss {step_loss:.8f}')
  progress.clear()
  progress.close()

  params = {
      'beta': lin_model.get_weights()[0],
      'beta_bias': lin_model.get_weights()[1],
  }
  eval_scores = eval_classifier(params, data_manager, eval_idxes)

  model_config = data_manager.db.get_metadata('model_config')
  linear_classifier = LinearClassifier(
      beta=params['beta'],
      beta_bias=params['beta_bias'],
      classes=data_manager.get_target_labels(),
      embedding_model_config=model_config,
  )
  return linear_classifier, eval_scores


@dataclasses.dataclass
class CsvWorkerState:
  """State for the CSV worker.

  Params:
    db: The base database from the parent thread.
    csv_filepath: The path to the CSV file to write.
    labels: The labels to write.
    threshold: The threshold for writing detections.
    _thread_db: The database to use in child threads.
  """

  db: db_interface.HopliteDBInterface
  csv_filepath: str
  labels: tuple[str, ...]
  threshold: float
  _thread_db: db_interface.HopliteDBInterface | None = None

  def get_thread_db(self) -> db_interface.HopliteDBInterface:
    if self._thread_db is None:
      self._thread_db = self.db.thread_split()
    return self._thread_db


def csv_worker_initializer(state: CsvWorkerState):
  """Initialize the CSV worker."""
  state.get_thread_db()
  with epath.Path(state.csv_filepath).open('w') as f:
    f.write('idx,project,filename,window_start,window_end,label,logits\n')


def csv_worker_fn(
    window_ids: np.ndarray, logits: np.ndarray, state: CsvWorkerState
) -> None:
  """Writes a CSV row for each detection.

  Args:
    window_ids: The window ids to write.
    logits: The logits for each window id.
    state: The state of the worker.
  """
  db = state.get_thread_db()
  with epath.Path(state.csv_filepath).open('a') as f:
    csv_writer = csv.writer(f)
    for idx, logit in zip(window_ids, logits):
      window = db.get_window(idx)
      recording = db.get_recording(window.recording_id)
      deployment = db.get_deployment(recording.deployment_id)
      for a in np.argwhere(logit > state.threshold):
        lbl = state.labels[a[0]]
        row = [
            idx,
            deployment.project,
            recording.filename,
            window.offsets[0],
            window.offsets[1],
            lbl,
            logit[a][0],
        ]
        row = [str(x) for x in row]
        csv_writer.writerow(row)


def batched_embedding_iterator(
    db: db_interface.HopliteDBInterface,
    window_ids: np.ndarray,
    batch_size: int = 1024,
) -> Iterator[tuple[np.ndarray, np.ndarray]]:
  """Iterate over embeddings in batches."""
  for q in range(0, len(window_ids), batch_size):
    batch_ids = window_ids[q : q + batch_size]
    batch_embs = db.get_embeddings_batch(batch_ids)
    yield batch_ids, batch_embs


def write_inference_csv(
    linear_classifier: LinearClassifier,
    db: db_interface.HopliteDBInterface,
    output_filepath: str,
    threshold: float,
    labels: Sequence[str] | None = None,
    window_ids: np.ndarray | Sequence[int] | None = None,
) -> None:
  """Write a CSV for all audio windows with logits above a threshold."""
  if window_ids is None:
    window_ids = np.array(db.match_window_ids())
  elif not isinstance(window_ids, np.ndarray):
    window_ids = np.array(window_ids)
  if labels is None:
    labels = linear_classifier.classes
  else:
    labels = tuple(set(labels).intersection(linear_classifier.classes))
  label_ids = {cl: i for i, cl in enumerate(linear_classifier.classes)}
  target_label_ids = np.array([label_ids[l] for l in labels])
  logits_fn = lambda batch_embs: linear_classifier(batch_embs)[
      :, target_label_ids
  ]
  detection_count = 0
  state = CsvWorkerState(
      db=db,
      csv_filepath=output_filepath,
      labels=labels,
      threshold=threshold,
  )
  emb_iter = batched_embedding_iterator(db, window_ids, batch_size=1024)
  with futures.ThreadPoolExecutor(
      max_workers=1,
      initializer=csv_worker_initializer,
      initargs=(state,),
  ) as executor:
    for batch_idxes, batch_embs in tqdm.tqdm(emb_iter):
      logits = logits_fn(batch_embs)
      # Filter out rows with no detections, avoiding extra database retrievals.
      detections = logits > threshold
      keep_rows = detections.max(axis=1)
      logits = logits[keep_rows]
      kept_idxes = batch_idxes[keep_rows]
      executor.submit(csv_worker_fn, kept_idxes, logits, state)
      detection_count += detections.sum()
  print(f'Wrote {detection_count} detections to {output_filepath}')



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for brute search functionality."""

import shutil
import tempfile

import numpy as np
from perch_hoplite.db import brutalism
from perch_hoplite.db import score_functions
from perch_hoplite.db.tests import test_utils

from absl.testing import absltest
from absl.testing import parameterized

EMBEDDING_SIZE = 8


class BrutalismTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.tempdir = tempfile.mkdtemp()

  def tearDown(self):
    super().tearDown()
    shutil.rmtree(self.tempdir)

  @parameterized.product(
      db_type=(
          'in_mem',
          'sqlite_usearch',
      ),
      sample_size=(
          None,
          0.5,
          128,
      ),
  )
  def test_threaded_brute_search(self, db_type, sample_size):
    rng = np.random.default_rng(42)
    db = test_utils.make_db(self.tempdir, db_type, 1000, rng, EMBEDDING_SIZE)
    query_idx = db.match_window_ids(limit=1)[0]
    query_embedding = db.get_embedding(query_idx)
    results = brutalism.brute_search(
        db,
        query_embedding,
        search_list_size=10,
        score_fn=score_functions.get_score_fn('cos'),
        sample_size=sample_size,
        rng_seed=42,
    )
    got_ids = [r.window_id for r in results]
    if sample_size is None:
      self.assertIn(query_idx, got_ids)
    self.assertLen(results.search_results, 10)

    # Check agreement of threaded brute search with the non-threaded version.
    t_results = brutalism.threaded_brute_search(
        db,
        query_embedding,
        search_list_size=10,
        batch_size=128,
        score_fn=score_functions.get_score_fn('cos'),
        sample_size=sample_size,
        rng_seed=42,
    )
    t_got_scores = [r.sort_score for r in t_results]
    got_scores = [r.sort_score for r in results]
    np.testing.assert_almost_equal(t_got_scores, got_scores)
    t_got_ids = [r.window_id for r in t_results]
    t_got_scores = [r.sort_score for r in t_results]
    num_equal_to_last = np.sum(np.isclose(t_got_scores, t_got_scores[-1]))
    self.assertSetEqual(
        # Embedding IDs need to be checked as a set due to score ties. Also, the
        # last score (and its corresponding embedding IDs) must be ignored due
        # to potential ties with results beyond top-k.
        set(got_ids[:-num_equal_to_last]),
        set(t_got_ids[:-num_equal_to_last]),
    )

  @parameterized.parameters(*test_utils.DB_TYPES)
  def test_get_random_embedding_scores(self, db_type):
    rng = np.random.default_rng(42)
    db = test_utils.make_db(self.tempdir, db_type, 1000, rng, EMBEDDING_SIZE)
    query_idx = db.match_window_ids(limit=1)[0]
    query_embedding = db.get_embedding(query_idx)
    scores = brutalism.get_random_embedding_scores(
        db,
        query_embedding,
        score_fn=score_functions.get_score_fn('cos'),
        sample_size=64,
        rng_seed=42,
    )
    self.assertEqual(scores.shape, (64,))


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functionality for embedding audio examples."""

from concurrent import futures
import dataclasses
import functools
import itertools
import threading

from absl import logging
import audioread
from ml_collections import config_dict
import numpy as np
from perch_hoplite import audio_io
from perch_hoplite.agile import metadata
from perch_hoplite.agile import source_info
from perch_hoplite.db import datatypes
from perch_hoplite.db import interface as hoplite_interface
from perch_hoplite.zoo import model_configs
from perch_hoplite.zoo import zoo_interface
import soundfile


@dataclasses.dataclass
class ModelConfig(datatypes.HopliteConfig):
  """Configuration for embedding model.

  Attributes:
    model_key: Key for the model wrapper class.
    embedding_dim: Dimensionality of the embedding.
    model_config: Config dict of arguments to instantiate the model wrapper.
    logit_key: If provided, model predictions will be stored instead of raw
      embeddings.
    logit_idxes: When storing model predictions, allows selecting a subset of
      prediction classes.
  """

  model_key: str
  embedding_dim: int
  model_config: config_dict.ConfigDict
  logits_key: str | None = None
  logits_idxes: tuple[int, ...] | None = None


def worker_initializer(state):
  name = threading.current_thread().name
  state[name + 'db'] = state['db'].thread_split()


def process_source_id(
    state,
    source_id: source_info.SourceId,
    window_size_s: float,
):
  """Process a single audio source."""
  worker = state['worker']
  # To access the thread-specific DB, we need to use the thread name.
  # name = threading.current_thread().name
  # db = state[name + 'db']
  glob = worker.audio_globs[source_id.dataset_name]
  target_sample_rate = worker.get_sample_rate_hz(source_id)
  audio_array = worker.load_audio(source_id)

  if audio_array is None:
    return
  if (
      audio_array.shape[0]
      < glob.min_audio_len_s * worker.embedding_model.sample_rate
  ):
    return

  outputs = worker.embedding_model.embed(audio_array)
  logits_key = state['worker'].model_config.logits_key
  if logits_key is None:
    embeddings = outputs.embeddings
  else:
    embeddings = outputs.logits[logits_key]
    logits_idxes = state['worker'].model_config.logits_idxes
    if logits_idxes is not None:
      embeddings = embeddings[..., logits_idxes]
    # Add channel axis to match the expected shape of the embeddings.
    embeddings = embeddings[:, np.newaxis, :]

  if embeddings is None:
    return

  sources = []
  offsets = []
  embs = []

  hop_size_s = worker.compute_hop_size_s(source_id, target_sample_rate)
  for t, embedding in enumerate(embeddings):
    offset_s = source_id.offset_s + t * hop_size_s
    offsets_list = [offset_s, offset_s + window_size_s]
    for channel_embedding in embedding:
      sources.append(source_id)
      offsets.append(offsets_list)
      embs.append(channel_embedding)
  return sources, offsets, embs


# TODO(tomdenton): Use itertools.batched in Python 3.12+
def batched(iterable, n):
  it = iter(iterable)
  while batch := tuple(itertools.islice(it, n)):
    yield batch


class EmbedWorker:
  """Worker for embedding audio examples."""

  def __init__(
      self,
      audio_sources: source_info.AudioSources,
      model_config: ModelConfig,
      db: hoplite_interface.HopliteDBInterface,
      embedding_model: zoo_interface.EmbeddingModel | None = None,
      audio_worker_threads: int = 8,
  ):
    self.db = db
    self.model_config = model_config
    self.audio_sources = audio_sources
    self.audio_worker_threads = audio_worker_threads
    if embedding_model is None:
      model_class = model_configs.get_model_class(model_config.model_key)
      self.embedding_model = model_class.from_config(model_config.model_config)
    else:
      self.embedding_model = embedding_model
    self.window_size_s = getattr(self.embedding_model, 'window_size_s')
    self.audio_globs = {
        g.dataset_name: g for g in self.audio_sources.audio_globs
    }
    self.metadata = {
        g.dataset_name: metadata.AgileMetadata.from_directory(g.base_path)
        for g in self.audio_sources.audio_globs
    }

  def _log_error(self, source_id, exception, counter_name):
    logging.warning(
        'The audio at (%s / %f) could not be loaded (%s). '
        'The exception was (%s)',
        source_id.filepath,
        source_id.offset_s,
        counter_name,
        exception,
    )

  def _update_audio_sources(self):
    """Validates the embed config and/or saves it to the DB."""
    db_metadata = self.db.get_metadata(None)
    if 'audio_sources' not in db_metadata:
      self.db.insert_metadata(
          'audio_sources', self.audio_sources.to_config_dict()
      )
      return

    db_audio_sources = source_info.AudioSources.from_config_dict(
        db_metadata['audio_sources']
    )
    merged = self.audio_sources.merge_update(db_audio_sources)
    self.db.insert_metadata('audio_sources', merged.to_config_dict())
    self.audio_sources = merged

  def _update_model_config(self):
    """Validates the model config and/or saves it to the DB."""
    db_metadata = self.db.get_metadata(None)
    if 'model_config' not in db_metadata:
      self.db.insert_metadata(
          'model_config', self.model_config.to_config_dict()
      )
      return

    db_model_config = ModelConfig(**db_metadata['model_config'])
    if self.model_config == db_model_config:
      return

    # Validate the config against the DB.
    # TODO(tomdenton): Implement compatibility checks for model configs.
    if self.model_config.model_key != db_model_config.model_key:
      raise AssertionError(
          'The configured model key does not match the model key that is '
          'already in the DB.'
      )
    if self.model_config.embedding_dim != db_model_config.embedding_dim:
      raise AssertionError(
          'The configured embedding dimension does not match the embedding '
          'dimension that is already in the DB.'
      )
    self.db.insert_metadata('model_config', self.model_config.to_config_dict())

  def update_configs(self):
    """Validates the configs and saves them to the DB."""
    self._update_model_config()
    self._update_audio_sources()
    self.db.commit()

  @functools.cache  # pylint: disable=method-cache-max-size-none
  def _get_or_insert_deployment_id(
      self,
      deployment_name: str,
      project_name: str,
      error_on_insert: bool = False,
  ) -> int:
    """Get the deployment ID for the given deployment name and project name."""
    deployments = self.db.get_all_deployments(
        config_dict.create(eq=dict(name=deployment_name, project=project_name))
    )
    if deployments:
      return deployments[0].id
    if error_on_insert:
      raise ValueError(
          f'Deployment {deployment_name} not found in project {project_name}.'
      )
    md = self.metadata[project_name].get_deployment_metadata(deployment_name)
    md.pop('deployment', None)
    return self.db.insert_deployment(
        name=deployment_name,
        project=project_name,
        **md,
    )

  @functools.cache  # pylint: disable=method-cache-max-size-none
  def _get_or_insert_recording_id(
      self,
      filename: str,
      deployment_id: int,
      dataset_name: str,
      error_on_insert: bool = False,
  ) -> tuple[int, bool]:
    """Get the recording ID, and indicate whether it was newly inserted."""
    recordings = self.db.get_all_recordings(
        config_dict.create(
            eq=dict(filename=filename, deployment_id=deployment_id)
        )
    )
    if recordings:
      return recordings[0].id, False
    if error_on_insert:
      raise ValueError(
          f'Recording {filename} not found in deployment {deployment_id}.'
      )
    md = self.metadata[dataset_name].get_recording_metadata(filename)
    md.pop('recording', None)
    return (
        self.db.insert_recording(
            filename=filename,
            deployment_id=deployment_id,
            **md,
        ),
        True,
    )

  def add_deployments(self, target_dataset_name: str | None = None):
    """Add deployments to db and create a source ID to deployment ID mapping."""
    # Create missing deployments in the database.
    for source in self.audio_sources.iterate_all_sources(target_dataset_name):
      deployment_name = source.deployment_name_from_file_id()
      self._get_or_insert_deployment_id(
          deployment_name=deployment_name,
          project_name=source.dataset_name,
      )
    self.db.commit()

  def add_recordings(self, target_dataset_name: str | None = None) -> set[int]:
    """Add recordings to db and create a source ID to recording ID mapping."""
    new_recordings = set([])
    for source in self.audio_sources.iterate_all_sources(target_dataset_name):
      deployment_id = self._get_or_insert_deployment_id(
          deployment_name=source.deployment_name_from_file_id(),
          project_name=source.dataset_name,
      )
      recording_id, is_new = self._get_or_insert_recording_id(
          source.file_id,
          deployment_id,
          source.dataset_name,
      )
      if is_new:
        new_recordings.add(recording_id)
    self.db.commit()
    return new_recordings

  def add_annotations(self, target_dataset_name: str | None = None):
    """Add annotations from metadata to db."""
    dataset_names = self.metadata.keys()
    if target_dataset_name is not None:
      dataset_names = [target_dataset_name]

    for dataset_name in dataset_names:
      if dataset_name not in self.metadata:
        continue
      agile_md = self.metadata[dataset_name]
      if not agile_md.annotations:
        continue
      for file_id, annotation_list in agile_md.annotations.items():
        depl_name = file_id.split('/')[0]
        if not depl_name:
          logging.warning(
              'Could not get deployment name from file_id %s, skipping.',
              file_id,
          )
          continue
        depl_id = self._get_or_insert_deployment_id(depl_name, dataset_name)
        rec_id, _ = self._get_or_insert_recording_id(
            file_id, depl_id, dataset_name
        )
        for annotation in annotation_list:
          self.db.insert_annotation(
              rec_id,
              annotation.offsets,
              annotation.label,
              annotation.label_type,
              provenance=annotation.provenance,
              handle_duplicates='skip',
          )
    self.db.commit()

  def embed_dataset(
      self,
      batch_size=32,
      handle_duplicates: str = 'error',
      target_dataset_name: str | None = None,
      new_recordings: set[int] | None = None,
  ):
    """Embed audio examples from the given dataset."""
    # Process all sources.
    state = {}
    state['db'] = self.db
    state['worker'] = self
    state['new_recordings'] = new_recordings
    with futures.ThreadPoolExecutor(
        max_workers=self.audio_worker_threads,
        initializer=worker_initializer,
        initargs=(state,),
    ) as executor:
      source_iterator = self.audio_sources.iterate_all_sources(
          target_dataset_name
      )
      for source_ids_batch in batched(source_iterator, batch_size):
        got = executor.map(
            process_source_id,
            itertools.repeat(state),
            source_ids_batch,
            itertools.repeat(self.window_size_s),
        )
        # TODO(tomdenton): Consider using a db writer thread to avoid blocking.
        for result in got:
          if result is None:
            continue
          recording_ids = []
          for s in result[0]:
            deployment_id = self._get_or_insert_deployment_id(
                s.deployment_name_from_file_id(), s.dataset_name
            )
            recording_id, _ = self._get_or_insert_recording_id(
                s.file_id, deployment_id, s.dataset_name
            )
            recording_ids.append(recording_id)
          if all(r in new_recordings for r in recording_ids):
            dupe_strategy = 'allow'
          else:
            dupe_strategy = handle_duplicates
          windows_batch = [
              {
                  'recording_id': recording_ids[i],
                  'offsets': o,
              }
              for i, (_, o, _) in enumerate(zip(*result))
          ]
          embeddings_batch = np.array(result[2])
          self.db.insert_windows_batch(
              windows_batch,
              embeddings_batch,
              handle_duplicates=dupe_strategy,
          )
    self.db.commit()

  def get_sample_rate_hz(self, source_id: source_info.SourceId) -> int:
    """Get the sample rate of the embedding model."""
    dataset_name = source_id.dataset_name
    if dataset_name not in self.audio_globs:
      raise ValueError(f'Dataset name {dataset_name} not found in audio globs.')
    audio_glob = self.audio_globs[dataset_name]
    if audio_glob.target_sample_rate_hz == -2:
      return self.embedding_model.sample_rate
    elif audio_glob.target_sample_rate_hz == -1:
      # Uses the file's native sample rate.
      return -1
    elif audio_glob.target_sample_rate_hz > 0:
      return audio_glob.target_sample_rate_hz
    else:
      raise ValueError('Invalid target_sample_rate.')

  def load_audio(self, source_id: source_info.SourceId) -> np.ndarray | None:
    """Load audio from the indicated source and log any problems."""
    target_sample_rate_hz = self.get_sample_rate_hz(source_id)
    try:
      audio_array = audio_io.load_audio_window(
          filepath=source_id.filepath,
          offset_s=source_id.offset_s,
          sample_rate=target_sample_rate_hz,
          window_size_s=source_id.shard_len_s,
      )
      return np.array(audio_array)
    except soundfile.LibsndfileError as inst:
      self._log_error(source_id, inst, 'audio_libsndfile_error')
    except ValueError as inst:
      self._log_error(source_id, inst, 'audio_bad_offset')
    except audioread.NoBackendError as inst:
      self._log_error(source_id, inst, 'audio_no_backend')
    except EOFError as inst:
      self._log_error(source_id, inst, 'audio_eof_error')
    except RuntimeError as inst:
      if 'Soundfile is not available' in str(inst):
        self._log_error(source_id, inst, 'audio_no_soundfile')
      else:
        self._log_error(source_id, inst, 'audio_runtime_error')

  def compute_hop_size_s(
      self,
      source_id: source_info.SourceId,
      target_sample_rate_hz: int,
      model_hop_size_s: float | None = None,
  ) -> float:
    """Compute the hop size of the embedding model."""

    if model_hop_size_s is not None:
      return model_hop_size_s
    if hasattr(self.embedding_model, 'hop_size_s'):
      model_hop_size_s = float(self.embedding_model.hop_size_s)
    else:
      # TODO(tomdenton): Allow user specified hop size.
      raise ValueError('hop_size_s is not defined for the model.')
    model_sample_rate = self.embedding_model.sample_rate

    if target_sample_rate_hz == -2:
      return model_hop_size_s
    elif target_sample_rate_hz == -1:
      audio_sample_rate = source_id.sample_rate_hz
    elif target_sample_rate_hz > 0:
      audio_sample_rate = target_sample_rate_hz
    else:
      raise ValueError('Invalid target_sample_rate.')
    return model_hop_size_s * model_sample_rate / audio_sample_rate

  def process_all(
      self,
      target_dataset_name: str | None = None,
      batch_size=32,
      handle_duplicates='error',
  ):
    """Process all audio examples."""

    # Update model config and audio sources in the database.
    self.update_configs()
    if self.db.count_embeddings() == 0:
      # No chance of duplicates, so we can use "allow" mode.
      handle_duplicates = 'allow'

    # Add deployments and recordings to the database.
    print('\nAdding deployments...')
    self.add_deployments(target_dataset_name)
    print('\nAdding recordings...')
    new_recordings = self.add_recordings(target_dataset_name)
    print('\nAdding annotations...')
    self.add_annotations()
    print('\nEmbedding audio...')
    self.embed_dataset(
        batch_size=batch_size,
        handle_duplicates=handle_duplicates,
        target_dataset_name=target_dataset_name,
        new_recordings=new_recordings,
    )
    self.db.commit()
